#!/usr/bin/perl -w
#
use strict;
use RNA;


my ($precursor_stru_file)=$ARGV[0];
my ($sorted_map_file)=$ARGV[1];

my %pre_info;
my $pre_name="";
my $mat_min_bpct=16;
my $star_min_bpct=16;
my $min_pre_len=55;
my $min_star_reads=1;
my $default_pre_anti_ratio=0.15;
my $default_fuzzy_ratio=0.25;
my $default_loop_ratio=0.1;

open (Stru,"$precursor_stru_file") || die;

while(<Stru>){
	if(/>(chr(\d+)\:(\d+)\_(\d+)([+-])\S+)/){
		$pre_name=$1;
		$pre_info{$pre_name}{"chr"}=$2;
		$pre_info{$pre_name}{"ln"}=$3;
		$pre_info{$pre_name}{"rn"}=$4;
		$pre_info{$pre_name}{"strand"}=$5;
	}
	if(/([A-Za-z]{100,})/){
		$pre_info{$pre_name}{"seq"}=$1;
	}
	if(/([\(\)\.]{100,})\s+/){
		$pre_info{$pre_name}{"stru"}=$1;
	}
}
close(Stru);



open (Tag,"$sorted_map_file") || die;

my $pre_subj_name="";
my $subj_name="";
my $mir_block_st=0;
my $mir_block_end=0;
my %tag_map=();
my %ct_bp=();
my %sorted_feature=();

while (<Tag>){
	if(/^(\S+\-(\d+))\s+([+-])\s+(\S+\:(\d+)\_(\d+)([+-])(\S+))\s+(\d+)\s+(\S+)/){
		my ($tag_name,$tag_num,$strand,$subj_st,$subj_end,$seg_strand,$up_down,$tag_st,$tag_seq)=($1,$2,$3,$5,$6,$7,$8,$9,$10);
		$subj_name=$4;
		if($pre_subj_name && $pre_subj_name ne $subj_name){
			mir_seq_parse();			
			$mir_block_end=0;
			%tag_map=();
			%ct_bp=();
			%sorted_feature=();
		}
		if(($up_down eq "up" && $seg_strand eq "+") || ($up_down eq "down" && $seg_strand eq "-")){
			$mir_block_st=400;
		}
		else{
			$mir_block_st=50;
		}
		$mir_block_end=$mir_block_st+($subj_end-$subj_st+1-450) if (!$mir_block_end);
		my $tag_len=length($tag_seq);
		$tag_map{$tag_name}{"freq"}=$tag_num;
		$tag_map{$tag_name}{"strand"}=$strand;
		$tag_map{$tag_name}{"st"}=$tag_st;
		$tag_map{$tag_name}{"end"}=$tag_st+$tag_len-1;
		$tag_map{$tag_name}{"info"}=$_;

		if((!$sorted_feature{"mat"}{"end"}) and ($mir_block_end>=$tag_map{$tag_name}{"end"}) and ($mir_block_st<=$tag_st)){
			$sorted_feature{"mat"}{"st"}=$tag_map{$tag_name}{"st"};
			$sorted_feature{"mat"}{"end"}=$tag_map{$tag_name}{"end"};
			$sorted_feature{"mat"}{"tag"}=$tag_name;
		}
		$pre_subj_name=$subj_name;
	}
}

mir_seq_parse();
close(Tag);

#--------------

sub stru_bp{
	my $stru=$pre_info{$pre_subj_name}{"stru"};
	my @stru_ary=split(//,$stru);
	my @pack;
	foreach my $pos (0..$#stru_ary){
		push (@pack,$pos+1) if ($stru_ary[$pos] eq "(");
		if($stru_ary[$pos] eq ")"){
			my $pre_pos=pop(@pack);
	    		$ct_bp{$pre_pos}=$pos+1;
	    		$ct_bp{$pos+1}=$pre_pos;
		}
	}
}

#--------------

sub mir_seq_parse{
	stru_bp();
	my $mat_bracket_ct=sub_stru_pattern("mat",$pre_subj_name);
	return 0 if ($mat_bracket_ct<$mat_min_bpct);
	my $star_bracket_ct=star_mir_region();
	return 0 if ($star_bracket_ct<$star_min_bpct);
	my $bplen_diff1=asym_ck("mat","star");
	return 0 if ($bplen_diff1>2);
	my $pre_len_ck=pre_location();
	return 0 if (!$pre_len_ck);
	my $hp_stru_ck=hairpin_stru();
	return 0 if (!$hp_stru_ck);
	my $bplen_diff2=asym_ck("pre_mat","pre_star");
	return 0 if ($bplen_diff2>2);
	loop_location();
	my $filter_ck=tag_freq_map();
	return 0 if (!$filter_ck);
	print "mature_bracket_ct:	$mat_bracket_ct\n";
	print "star_bracket_ct:	$star_bracket_ct\n\n";
	mir_print();
}

#--------------

sub mat_mir_region{
	my ($st,$end,$freq)=(0,0,0);
	my @tag=reverse sort {$tag_map{$a}{"freq"} <=> $tag_map{$b}{"freq"}} keys %tag_map;
	foreach my $qry (@tag){
		if($tag_map{$qry}{"st"}>=$mir_block_st && $tag_map{$qry}{"end"}<=$mir_block_end){
			$sorted_feature{"mat"}{"st"}=$tag_map{$qry}{"st"};
			$sorted_feature{"mat"}{"end"}=$tag_map{$qry}{"end"};
			$sorted_feature{"mat"}{"tag"}=$qry;
			last;
		}
	}
	return;
}

#--------------

sub sub_stru_pattern{
	my ($item,$host_name)=@_;
	my ($host_seq,$host_stru)=("","");
	if($host_name eq "pre"){
		$host_seq=$sorted_feature{$host_name}{"seq"};
		$host_stru=$sorted_feature{$host_name}{"stru"};
	}
	else{
		$host_seq=$pre_info{$host_name}{"seq"};
		$host_stru=$pre_info{$host_name}{"stru"};
	}
	my $sub_stru=substr($host_stru,$sorted_feature{$item}{"st"}-1,$sorted_feature{$item}{"end"}-$sorted_feature{$item}{"st"}+1);
	return 0 if ($sub_stru=~/\(.*\)/ || $sub_stru=~/\).*\(/ || $sub_stru=~/\.\.\.\./);
	my $bracket_ct=0;
	my $prime="";
	$sorted_feature{$item}{"seq"}=substr($host_seq,$sorted_feature{$item}{"st"}-1,$sorted_feature{$item}{"end"}-$sorted_feature{$item}{"st"}+1);
	$sorted_feature{$item}{"stru"}=$sub_stru;
	while($sub_stru=~/\(/g){
		$bracket_ct++;
		$sorted_feature{$item}{"prime"}="5prime" if ($item=~/mat/);
	}
	while($sub_stru=~/\)/g){
		$bracket_ct++;
		$sorted_feature{$item}{"prime"}="3prime" if ($item=~/mat/);
	}
	return ($bracket_ct);
}

#--------------

sub star_mir_region{
	my $mat_st=$sorted_feature{"mat"}{"st"};
	my $mat_end=$sorted_feature{"mat"}{"end"};
	my ($star_st,$st_offset,$star_end,$end_offset)=(0,0,0,0);
	while(!$star_st){
		if($ct_bp{$mat_end-2-$st_offset}){
			$star_st=$ct_bp{$mat_end-2-$st_offset}-$st_offset;
			last if (!$star_st);
		}
		else{
			$st_offset++;
		}
	}
	while(!$star_end){
		if($ct_bp{$mat_st+$end_offset}){
			$star_end=$ct_bp{$mat_st+$end_offset}+$end_offset+2;
		}
		else{
			$end_offset++;
		}
	}
	$sorted_feature{"star"}{"st"}=$star_st>0 ? $star_st : 1;
	$sorted_feature{"star"}{"end"}=$star_end<=length($pre_info{$pre_subj_name}{"seq"}) ? $star_end : length($pre_info{$pre_subj_name}{"seq"});
	return 0 if ($star_st<0 || $star_end>($mir_block_end-$mir_block_st+1+450) || ($mat_end>$star_st && $star_end>$mat_st));
	my $bracket_ct=sub_stru_pattern("star",$pre_subj_name);
	return ($bracket_ct);
}

#--------------

sub pre_location{
	my $pre_st=$sorted_feature{"mat"}{"st"}-2;
	my $pre_end=$sorted_feature{"star"}{"end"};
	my $pre_len=$pre_end-$pre_st+1;
	return 0 if ($pre_len<=$min_pre_len);
	if ($sorted_feature{"mat"}{"prime"} eq "3prime"){
		$pre_st=$sorted_feature{"star"}{"st"};
		$pre_end=$sorted_feature{"mat"}{"end"}+2;
	}
	my $up_st=$pre_st-1>10 ? $pre_st-11 : 0;
	my $down_len=length($pre_info{$pre_subj_name}{"stru"})-$pre_end>10 ? 10 : length($pre_info{$pre_subj_name}{"stru"})-$pre_end;
	my $up_flank_stru=substr($pre_info{$pre_subj_name}{"stru"},$up_st,$pre_st-1-$up_st);
	my $down_flank_stru=substr($pre_info{$pre_subj_name}{"stru"},$pre_end,$down_len);
	my ($up_stem_extend,$down_stem_extend)=(0,0);
	if($up_flank_stru=~/([\.|\(]*)$/){
		$up_stem_extend=length($1);
	}
	if($down_flank_stru=~/^([\.|\)]*)/){
		$down_stem_extend=length($1);
	}
	$pre_st-=$up_stem_extend;
	$pre_end+=$down_stem_extend;
	$sorted_feature{"pre"}{"st"}=$pre_st;
	$sorted_feature{"pre"}{"end"}=$pre_end;
	return 1;
}

#--------------

sub hairpin_stru{
	my $pre_seq=substr($pre_info{$pre_subj_name}{"seq"},$sorted_feature{"pre"}{"st"}-1,$sorted_feature{"pre"}{"end"}-$sorted_feature{"pre"}{"st"}+1);
	my ($pre_stru,$pre_mfe)=RNA::fold($pre_seq);
	$sorted_feature{"pre_mat"}{"st"}=$sorted_feature{"mat"}{"st"}-$sorted_feature{"pre"}{"st"}+1;
	$sorted_feature{"pre_mat"}{"end"}=$sorted_feature{"mat"}{"end"}-$sorted_feature{"pre"}{"st"}+1;
	$sorted_feature{"pre_star"}{"st"}=$sorted_feature{"star"}{"st"}-$sorted_feature{"pre"}{"st"}+1;
	$sorted_feature{"pre_star"}{"end"}=$sorted_feature{"star"}{"end"}-$sorted_feature{"pre"}{"st"}+1;
	$sorted_feature{"pre"}{"stru"}=$pre_stru;
	$sorted_feature{"pre"}{"mfe"}=$pre_mfe;
	$sorted_feature{"pre"}{"seq"}=$pre_seq;
	my $pre_mat_bracket_ct=sub_stru_pattern("pre_mat","pre");
	my $pre_star_bracket_ct=sub_stru_pattern("pre_star","pre");
	if($pre_mat_bracket_ct<$mat_min_bpct || $pre_star_bracket_ct<$star_min_bpct || ($sorted_feature{"mat"}{"prime"} ne $sorted_feature{"pre_mat"}{"prime"})){
		return 0;
	}
	return 1;
}

#--------------

sub loop_location{
	my ($loop_st,$loop_end)=(0,0);
	if($sorted_feature{"mat"}{"prime"} eq "5prime"){
		$loop_st=$sorted_feature{"mat"}{"end"}+1;
		$loop_end=$sorted_feature{"star"}{"st"}-1;		
	}
	else{
		$loop_st=$sorted_feature{"star"}{"end"}+1;
		$loop_end=$sorted_feature{"mat"}{"st"}-1;		
	}
	$sorted_feature{"loop"}{"st"}=$loop_st;
	$sorted_feature{"loop"}{"end"}=$loop_end;
	$sorted_feature{"loop"}{"seq"}=substr($pre_info{$pre_subj_name}{"seq"},$sorted_feature{"loop"}{"st"}-1,$sorted_feature{"loop"}{"end"}-$sorted_feature{"loop"}{"st"}+1);
	$sorted_feature{"loop"}{"stru"}=substr($pre_info{$pre_subj_name}{"stru"},$sorted_feature{"loop"}{"st"}-1,$sorted_feature{"loop"}{"end"}-$sorted_feature{"loop"}{"st"}+1);
}

#--------------

sub asym_ck{
	my ($name1,$name2)=@_;
	my $len1=length($sorted_feature{$name1}{"seq"});
	my $len2=length($sorted_feature{$name2}{"seq"});
	my $len_diff=abs($len1-$len2);
	return $len_diff;
}

#--------------

sub tag_freq_map{
	my ($pre_anti_freq,$pre_sense_freq,$mat_freq,$mat_fuzzy_freq,$star_freq,$real_star_freq,$star_fuzzy_freq,$loop_freq)=(0,0,0,0,0,0,0,0);
	my ($star_tag,$real_star_tag,$loop_tag,$pre_tag)=("","","","");
	my @tag=sort {$tag_map{$a}{"st"} <=> $tag_map{$b}{"st"}} keys %tag_map;
	foreach my $qry (@tag){
		last if ($tag_map{$qry}{"st"}>$sorted_feature{"pre"}{"end"});
		if(overlap("pre",$qry)){
			$pre_tag.=$qry . ";";
			if($tag_map{$qry}{"strand"} eq "+"){
				$pre_sense_freq+=$tag_map{$qry}{"freq"};
			}
			else{
				$pre_anti_freq+=$tag_map{$qry}{"freq"};
			}
		}		
		next if ($tag_map{$qry}{"strand"} eq "-");
		if(contain("mat",$qry)){
			$mat_freq+=$tag_map{$qry}{"freq"};
		}		
		if(overlap("mat",$qry) && !contain("mat",$qry)){
			$mat_fuzzy_freq+=$tag_map{$qry}{"freq"};
		}
		if(overlap("star",$qry) && !contain("star",$qry)){
			$star_fuzzy_freq+=$tag_map{$qry}{"freq"};
		}
		if(contain("star",$qry)){
			$star_freq+=$tag_map{$qry}{"freq"};
			$star_tag.=$qry . ";";
		}
		if(contain("star",$qry) && star_tag($qry)){
			$real_star_freq+=$tag_map{$qry}{"freq"};
			$real_star_tag.=$qry . ";";			
		}
		if(contain("loop",$qry)){
			$loop_freq+=$tag_map{$qry}{"freq"};
			$loop_tag.=$qry . ";";
		}
	}
	return 0 if ($star_freq+$mat_freq==0);
	my $pre_anti_ratio=$pre_anti_freq/$pre_sense_freq;
	my $fuzzy_ratio=($mat_fuzzy_freq+$star_fuzzy_freq)/($star_freq+$mat_freq);
	my $loop_ratio=$loop_freq/($star_freq+$mat_freq);
	return 0 if (($pre_anti_ratio>$default_pre_anti_ratio) || ($fuzzy_ratio>$default_fuzzy_ratio) || ($loop_ratio>$default_loop_ratio) || ($real_star_freq<$min_star_reads)); ###set freq filter here
	my $cur_rand_pval=randfold_cal();
	return 0 if ($cur_rand_pval>0.05);
	print "\n#############################\n\n";
	print "pre_anti_freq:	$pre_anti_freq\n";
	print "pre_sense_freq:	$pre_sense_freq\n";
	print "mat_fuzzy_freq:	$mat_fuzzy_freq\n";
	print "star_fuzzy_freq:	$star_fuzzy_freq\n";
	print "pre_anti_ratio:	$pre_anti_ratio\n";
	print "fuzzy_ratio:	$fuzzy_ratio\n";
	print "loop_ratio:	$loop_ratio\n";
	print "randfold pval:	$cur_rand_pval\n";
	$sorted_feature{"mat"}{"freq"}=$mat_freq;
	$sorted_feature{"star"}{"tag"}=$star_tag;	
	$sorted_feature{"realstar"}{"tag"}=$real_star_tag;
	$sorted_feature{"loop"}{"tag"}=$loop_tag;	
	$sorted_feature{"pre"}{"tag"}=$pre_tag;	
	$sorted_feature{"star"}{"freq"}=$star_freq;
	$sorted_feature{"realstar"}{"freq"}=$real_star_freq;
	$sorted_feature{"loop"}{"freq"}=$loop_freq;	
	$sorted_feature{"pre"}{"freq"}=$pre_sense_freq;	
	return 1;
}

#--------------

sub overlap{
	my ($feature,$qry)=@_;
	my $st=$sorted_feature{$feature}{"st"};
	my $end=$sorted_feature{$feature}{"end"};
	if($feature eq "mat"){
		$st=$sorted_feature{$feature}{"st"};
		$end=$sorted_feature{$feature}{"end"}-2;
	}
	if($feature eq "star"){
		$st=$sorted_feature{$feature}{"st"}+2;
		$end=$sorted_feature{$feature}{"end"}-2;		
	}	
	if(($tag_map{$qry}{"end"}>=$st) && ($end>=$tag_map{$qry}{"st"})){
		return 1;
	}
	else{
		return 0;
	}
}

#--------------

sub star_tag{
	my ($qry)=@_;
	my $offset=abs($sorted_feature{"star"}{"st"}-$tag_map{$qry}{"st"});
	if($offset<=1){
		return 1;
	}
	else{
		return 0;
	}
}

#--------------

sub contain{
	my ($feature,$qry)=@_;
	my ($st,$end)=(0,0);
	if($feature eq "mat" or $feature eq "star"){
		$st=$sorted_feature{$feature}{"st"}-2;
		$end=$sorted_feature{$feature}{"end"}+3;
	}
	else{
		$st=$sorted_feature{$feature}{"st"}-2;
		$end=$sorted_feature{$feature}{"end"}+2;		
	}
	if(($tag_map{$qry}{"st"}>=$st) && ($end>=$tag_map{$qry}{"end"})){
		return 1;
	}
	else{
		return 0;
	}
}

sub randfold_cal{
	my $seq=$1;
	open(Tmp,">tmp.fas") || die;
	print Tmp ">tmp\n",$sorted_feature{"pre"}{"seq"},"\n";
	close(Tmp);
	my $p_value=`randfold -d tmp.fas 999 | cut -f 3`;
	return ($p_value);
}

#--------------

sub mir_print{
	print "===============\n";
	print $pre_subj_name,"\n";
	print "===============\n";
	print "pre_st:	",$sorted_feature{"pre"}{"st"},"\n";
	print "pre_end:	",$sorted_feature{"pre"}{"end"},"\n";
	print "pre_seq:	",$sorted_feature{"pre"}{"seq"},"\n";
	print "pre_stru:	",$sorted_feature{"pre"}{"stru"},"\n";
	print "mat_seq:	",$sorted_feature{"mat"}{"seq"},"\n";
	print "mat_stru:	",$sorted_feature{"mat"}{"stru"},"\n";
	print "mat_pos:	",$sorted_feature{"pre_mat"}{"st"},"	",$sorted_feature{"pre_mat"}{"end"},"\n";
	print "mat_tag:	",$sorted_feature{"mat"}{"tag"},"\n";
	print "mat_freq:	",$sorted_feature{"mat"}{"freq"},"\n";
	print "star_seq:	",$sorted_feature{"star"}{"seq"},"\n";
	print "star_stru:	",$sorted_feature{"star"}{"stru"},"\n";
	print "star_pos:	",$sorted_feature{"pre_star"}{"st"},"	",$sorted_feature{"pre_star"}{"end"},"\n";
	print "real_star_tag:	",$sorted_feature{"realstar"}{"tag"},"\n";
	print "real_star_freq:	",$sorted_feature{"realstar"}{"freq"},"\n";	
	print "all_star_tag:	",$sorted_feature{"star"}{"tag"},"\n";
	print "all_star_freq:	",$sorted_feature{"star"}{"freq"},"\n";
	print "loop_seq:	",$sorted_feature{"loop"}{"seq"},"\n";
	print "loop_stru:	",$sorted_feature{"loop"}{"stru"},"\n";
	print "loop_pos:	",$sorted_feature{"loop"}{"st"}-$sorted_feature{"pre"}{"st"}+1,"	",$sorted_feature{"loop"}{"end"}-$sorted_feature{"pre"}{"st"}+1,"\n";
	print "loop_tag:	",$sorted_feature{"loop"}{"tag"},"\n";
	print "loop_freq:	",$sorted_feature{"loop"}{"freq"},"\n";
	print "mapped tags:\n";	
	my @tag=reverse sort {$tag_map{$a}{"freq"} <=> $tag_map{$b}{"freq"}} keys %tag_map;
	foreach my $qry (@tag){
		if ($sorted_feature{"pre"}{"tag"}=~/$qry\;/){
			if ($tag_map{$qry}{"info"}=~/\s+(\d+)\s+/){
				my $tag_loc=$1;
				my $mod_tag_loc=$tag_loc-$sorted_feature{"pre"}{"st"}+1;
				$tag_map{$qry}{"info"}=~s/\s+$tag_loc\s+/	$mod_tag_loc	/;
				print $tag_map{$qry}{"info"};
			}
		}
	}	
	print "\n";
}
