#!/usr/bin/perl -w
#
use strict;
use RNA;

my ($genome_file)=$ARGV[0];
my ($reads_file)=$ARGV[1];

my $chrm="";
my %chrm_seq;
my $seq_extend_up=400;
my $seq_extend_down=50;

open(Seq,"$genome_file") || die;	
while(<Seq>){
	if(/>chr[0]?(\d+)/){
		$chrm=$1;
		next;
	}
	if(/(\S+)/){
		my $cur_seq=$1;
		$chrm_seq{$chrm}{"seq"}.=uc($cur_seq);
		$chrm_seq{$chrm}{"len"}+=length($cur_seq);
	}
}
close(Seq);

open (Seqout,">precurser.fas");
open (Struout,">structure.data");
foreach my $chrm (1..12){
	my %chrm_map;
	open (Tag,"$reads_file") || die;
	while(<Tag>){
		if(/^\S+\s+([+-])\s+chr$chrm\s+(\d+)\s+(\S+)/){
			my ($strand,$g_st,$tag)=($1,$2,$3);
			my $g_end=$g_st+length($tag)-1;
			foreach my $i ($g_st..$g_end){
				$chrm_map{$strand}{$i}++;
			}
		}
	}
	close(Tag);
	foreach my $direct (keys %chrm_map){
		my ($block_st,$block_end)=(0,0);
		foreach my $loc (1..$chrm_seq{$chrm}{"len"}){
			$block_st=$loc if ($chrm_map{$direct}{$loc} && !$chrm_map{$direct}{$loc-1});
			if($chrm_map{$direct}{$loc} && !$chrm_map{$direct}{$loc+1}){
				$block_end=$loc;
				my $pre_up_seq=seq_sub($chrm,$block_st-$seq_extend_up,$block_end+$seq_extend_down,$direct);
				my ($up_structure,$up_mfe)=RNA::fold($pre_up_seq);
				print Seqout ">chr$chrm:",$block_st-$seq_extend_up,"_",$block_end+$seq_extend_down,$direct,"up\n";
				print Struout ">chr$chrm:",$block_st-$seq_extend_up,"_",$block_end+$seq_extend_down,$direct,"up\n";
				print Seqout "$pre_up_seq\n";
				$pre_up_seq=~s/T/U/g;
				print Struout "$pre_up_seq\n";
				print Struout "$up_structure ($up_mfe)\n";
				#-------
				my $pre_down_seq=seq_sub($chrm,$block_st-$seq_extend_down,$block_end+$seq_extend_up,$direct);
				my ($down_structure,$down_mfe)=RNA::fold($pre_down_seq);
				print Seqout ">chr$chrm:",$block_st-$seq_extend_down,"_",$block_end+$seq_extend_up,$direct,"down\n";
				print Struout ">chr$chrm:",$block_st-$seq_extend_down,"_",$block_end+$seq_extend_up,$direct,"down\n";
				print Seqout "$pre_down_seq\n";
				$pre_down_seq=~s/T/U/g;
				print Struout "$pre_down_seq\n";
				print Struout "$down_structure ($down_mfe)\n";				
			}
		}
	}
}


#--------------------

sub seq_sub{
	my ($chrm,$st,$end,$strand)=@_;
	my $seq=substr($chrm_seq{$chrm}{"seq"},$st-1,$end-$st+1);
	$seq=revcompl($seq) if ($strand eq "-");
	return ($seq);
}

#--------------------

sub revcompl{
	my ($s)=@_;
	$s=reverse($s);
	$s=~tr/atgcATGCU/tacgTACGA/;
	return($s);
}

#--------------------
